#include "Loader.hpp"

#include "Encoding.hpp"
#include "Error.hpp"
#include "Linker.hpp"

#include <string.h>
#include <stdlib.h>

#ifdef WIN32
	#include <windows.h>
#elif __APPLE__
#include <sys/mman.h>
#include <errno.h>
#elif __unix__
#include <sys/mman.h>
#endif

#include "../algnmalloc.h"
#include "../ptrint_t.h"

namespace SoftWire
{
	Loader::Loader(const Linker &linker, bool x64) : linker(linker), x64(x64)
	{
		MachineCode = 0;
		instructions = 0;
		listing = 0;
		possession = true;

		reset();
	}

	Loader::~Loader()
	{
		if(possession)
		{
			_aligned_free(MachineCode);
			MachineCode = 0;
		}

		delete instructions;
		instructions = 0;

		delete[] listing;
		listing = 0;
	}

	void (*Loader::callable(const char *entryLabel))()
	{
		if(finalized) throw Error("Cannot retrieve callable from finalized code");

		if(!MachineCode)
		{
			loadCode();
		}

		if(!entryLabel)
		{
			return (void(*)())MachineCode;
		}

		const unsigned char *entryPoint = resolveLocal(entryLabel, 0);

		if(!entryPoint)
		{
			throw Error("Entry point '%s' not found", entryLabel);
		}

		return (void(*)())entryPoint;
	}

	void (*Loader::finalize(const char *entryLabel))()
	{
		if(!MachineCode)
		{
			loadCode();
		}

		finalized = true;

		delete instructions;
		instructions = 0;

		delete[] listing;
		listing = 0;

		if(!entryLabel)
		{
			return (void(*)())MachineCode;
		}

		const unsigned char *entryPoint = resolveLocal(entryLabel, 0);

		if(!entryPoint)
		{
			throw Error("Entry point '%s' not found", entryLabel);
		}

		return (void(*)())entryPoint;
	}

	void *Loader::acquire()
	{
		possession = false;

		return MachineCode;
	}

	Encoding *Loader::appendEncoding(const Encoding &encoding)
	{
		if(!instructions) instructions = new Instruction();

		instructions->append(encoding);

		return instructions->tail();
	}

	void Loader::loadCode(const char *entryLabel)
	{
		int length = GetCodeSize(); //codeLength() + 64;   // NOTE: Code length is not accurate due to alignment issues

//		MachineCode = new unsigned char[length];

		// aligned code, zero pad,  by vesa

		#ifdef WIN32
		MachineCode =(unsigned char*)_aligned_malloc(length, 32);
        unsigned long oldProtection;
        VirtualProtect(MachineCode, length, PAGE_EXECUTE_READWRITE, &oldProtection);
		#elif __unix__
        MachineCode = (unsigned char *)malloc(length);
        if (mprotect(MachineCode, length, PROT_READ | PROT_WRITE | PROT_EXEC))
        {
            switch(errno)
            {
                case EINVAL:throw Error("JIT Compiler requires memory to be page-aligned (%x)",(unsigned)MachineCode);
                case EACCES:throw Error("Insufficient access permissions for code generation");
                case ENOTSUP:throw Error("JIT Compiler memory protection error");
                default:throw Error("Unknown memory protection error");
            }
        }
        #elif defined(__APPLE__)
        MachineCode = (unsigned char *)valloc(length);
        if (mprotect(MachineCode, length, PROT_READ | PROT_WRITE | PROT_EXEC))
        {
            switch(errno)
            {
                case EINVAL:throw Error("JIT Compiler requires memory to be page-aligned (%x)",(unsigned)MachineCode);
                case EACCES:throw Error("Insufficient access permissions for code generation");
                case ENOTSUP:throw Error("JIT Compiler memory protection error");
                default:throw Error("Unknown memory protection error");
            }
        }
        #else
        #error Please define memory protection
		#endif

		memset(MachineCode,0,length);
		unsigned char *currentCode = MachineCode;
		Instruction *instruction = instructions;

		while(instruction)
		{
			Encoding &encoding = *instruction;
			encoding.setAddress(currentCode);
			const char *reference = encoding.getReference();
			//const char *label = encoding.getLabel();

			if(reference)
			{
				if(encoding.relativeReference())
				{
					int offset = resolveReference(reference, instruction) - currentCode - encoding.length(currentCode);
					encoding.setJumpOffset(offset);
				}
				else
				{
					ptrint_t address = (ptrint_t)resolveReference(reference, instruction);

					// Encoded as memory reference or immediate?
					if(encoding.hasDisplacement())
					{
						encoding.addDisplacement(address);
					}
					else if(encoding.hasImmediate())
					{
						encoding.setImmediate(address);
					}
					else
					{
						INTERNAL_ERROR;
					}
				}
			}
			else if(encoding.hasImmediate() && encoding.relativeReference())
			{
				__int64 offset = encoding.getImmediate() - (__int64)currentCode - encoding.length(currentCode);
				encoding.setCallOffset((int)offset);
			}

			if(x64 && encoding.isRipRelative())
			{
				__int64 displacement = encoding.getDisplacement() - (__int64)currentCode - encoding.length(currentCode);
				encoding.setDisplacement((int)displacement);
			}

			currentCode += encoding.writeCode(currentCode);

			instruction = instruction->next();
		}
	}

	const unsigned char *Loader::resolveReference(const char *name, const Instruction *position) const
	{
		const unsigned char *reference = resolveLocal(name, position);
		
		if(reference)
		{
			return reference;
		}
		else
		{
			reference = resolveExternal(name);
		}

		if(reference)
		{
			return reference;
		}
		else
		{
			throw Error("Unresolved identifier '%s'", name);
		}
	}

	const unsigned char *Loader::resolveLocal(const char *name, const Instruction *position) const
	{
		const Instruction *instruction = instructions;

		unsigned char *target = MachineCode;
		unsigned char *local = 0;   // Closest match before, or after position

		while(instruction)
		{
			if(local && (instruction == position || position == 0))
			{
				return local;
			}
			else if(instruction == position)
			{
				position = 0;   // Return first match following position
			}

			if(instruction->getLabel() && strcmp(instruction->getLabel(), name) == 0)
			{
				local = target;
			}
			else
			{
				target += instruction->length(target);
			}

			instruction = instruction->next();
		}

		if(local)
		{
			return local;
		}

		return 0;
	}

	const unsigned char *Loader::resolveExternal(const char *name) const
	{
		return (unsigned char*)linker.resolveExternal(name);
	}

	int Loader::codeLength() const
	{
		const Instruction *instruction = instructions;
		unsigned char *end = MachineCode;

		while(instruction)
		{
			end += instruction->length(end);

			instruction = instruction->next();
		}

		return end - MachineCode;
	}

	const char *Loader::getListing()
	{
		if(!MachineCode)
		{
			return 0;
		}

		if(listing)
		{
			return listing;
		}

		listing = new char[codeLength() * 4];
		char *buffer = listing;

		for(Instruction *instruction = instructions; instruction; instruction = instruction->next())
		{
			buffer += instruction->printCode(buffer);
		}

		*(buffer - 1) = '\0';

		return listing;
	}

	void Loader::clearListing()
	{
		delete[] listing;
		listing = 0;
	}

	void Loader::reset()
	{
		if(possession)
		{
			if (MachineCode)
			{
				_aligned_free(MachineCode);
				MachineCode = 0;
			}
		}

		MachineCode = 0;

		delete instructions;
		instructions = 0;

		delete[] listing;
		listing = 0;

		possession = true;
		finalized = false;
	}

	int Loader::instructionCount()
	{
		Instruction *instruction = instructions;

		int count = 0;

		while(instruction)
		{
			if(instruction->isEmitting())
			{
				count++;
			}
			
			instruction = instruction->next();
		}

		return count;
	}
}
